#include <io.h>
#include <stdlib.h>
#include <string.h>
#include <dev/uart.h>
#include <dev/usartavr.h>
#include <sys/timer.h>
#include <terminal/btn-terminal.h>
#include <hardware/ram_banks.h>
#include <cdist/xbankdata.h>
#include <debug/logging.h>
#include <debug/toolbox.h>
#include <cm/con_mgr.h>
#include <cdist/hexfile.h>
#include <cdist/cdist-debug.h>

#define LOG_LEVEL 0
#define LOG_CLASS 7
#include <debug/log_set.h>

#define MIN_HEX_LINE_LEN (1+2+4+2+2 +1)    // incl. newline
#define MAX_HEX_LINE_LEN (1+2+4+2+32+2 +1) // incl. newline
#define HEX2BYTE(c) (u_char)(((c)<='9') ? (c)-'0' : (c) - 'A' + 10)

// errors
#define HEX_ERR_READ     1
#define HEX_ERR_FORMAT   2

#define MAX_DATA_LEN 16
typedef struct hex_record_s {
    u_char datalen;
    u_short addr;
    u_char type;
    u_char data[MAX_DATA_LEN];
    u_char cc;
} hex_record_t;

typedef struct hexfile_stack_s {
    FILE* stream;
    hex_record_t record;
    u_char uart_errors;
} hexfile_stack_t;

static hexfile_stack_t _hexfile_stack;

u_char _hex2byte(u_char *hex)
{
	u_char b = HEX2BYTE(hex[0]) << 4 | HEX2BYTE(hex[1]);
	return b;
}

u_char _read_fully(u_char *buffer, u_char count){
    short read;
    while(count > 0){
        read = _read(_fileno(stdout), buffer, count);
        if(read == -1)
            return 1;
        count -= (u_char) read;
        buffer += (u_char) read;
    }
    // check driver status
    _hexfile_stack.uart_errors = get_uart_errors(_hexfile_stack.stream);
    if(_hexfile_stack.uart_errors)
        return 1;
    return 0;
}

u_char _hex_record_read(void)
{
	u_char b, cc;
	u_char i;
	u_char linebuf[MAX_HEX_LINE_LEN];
	u_char *line = linebuf;

	// read static part
    if(_read_fully(linebuf, 9)){
        return HEX_ERR_READ;
    }

	if(*line != ':' )
		return HEX_ERR_FORMAT;
	line++;
		
	// checksum init
	cc = 0;
	
	// data length
	b = _hex2byte(line);
	if(b > MAX_DATA_LEN)
		return HEX_ERR_FORMAT;
	_hexfile_stack.record.datalen = b;
	cc += b;
	line+=2;
		
	// addr
	// high byte
	b = _hex2byte(line);
	_hexfile_stack.record.addr = b << 8;
	cc += b;
	line+=2;
	// low byte
	b = _hex2byte(line);
	_hexfile_stack.record.addr |= b;
	cc += b;
	line+=2;
	
	// record type
	b = _hex2byte(line);
	_hexfile_stack.record.type = b;
	cc += b;
	line+=2;

    // read variable part
    if(_read_fully(line, _hexfile_stack.record.datalen*2+2+1)){
        return HEX_ERR_READ;
    }
	
	// read data
	for(i = 0; i < _hexfile_stack.record.datalen; i++){
		b = _hex2byte(line);
		_hexfile_stack.record.data[i] = b;
		cc += b;
		line+=2;
	}
	
	// read cc field
	b = _hex2byte(line);
	_hexfile_stack.record.cc = b;
	
	// check cc
	cc = (u_char) ~cc + 1; // two's complement
	if(_hexfile_stack.record.cc != cc)
		return HEX_ERR_FORMAT;

	return 0;
}

void _loadhex_cmd_usage(void)
{
    tprintf("loadhex: usage: loadhex <ver> <type: 0=dsn, 1=tg> [<name>]\n");
}

void _loadhex_cmd(char * arg)
{
	u_char terminate, start_new_record;
	u_long base_addr, seg_start_addr;
	u_short lines;
    
    // parse cmd line and set program properties
    xbank_prog_info_t prog_info;
    
    
    char* carg = arg;
    prog_info.version = strtoul(carg, &carg, 10);
    if(arg == (char*) carg){
        _loadhex_cmd_usage();
        return;
    }
    arg = carg;
    prog_info.type = strtol(carg, &carg, 10);
    if(arg == (char*) carg){
        _loadhex_cmd_usage();
        return;
    }
    if(*carg != '\0')
        carg++; // omit space
#ifndef __BTN_UNIX__
    strlcpy(prog_info.name, carg, sizeof(prog_info.name));
#else
    strncpy(prog_info.name, carg, sizeof(prog_info.name) -1);
    prog_info.name[sizeof(prog_info.name) -1] = (u_char) 0;
#endif
    
    DEBUG("version: %lu\n", prog_info.version);
    DEBUG("type: %u\n", prog_info.type);
    DEBUG_STR("name: %s\n", prog_info.name);
    
    // get program lock
    if(xbank_get_program_lock()){
        tprintf(":LH failed: program locked\n");
        return; 
    }
	
	xbank_invalidate_program();
    xbank_set_prog_info(&prog_info);
	xbank_init_prog_write();
    
	// init 
	base_addr = 0;
	seg_start_addr = 0;
	start_new_record = 1;
	terminate = 0;
	lines = 0;
	
	con_mgr_inq_disable();
    NutSleep(1000);
	
	tprintf("ready to receive hexfile, press enter to quit.\n");
	while(!terminate){
		u_char retval;
        
		if((retval = _hex_record_read())){
            if(retval == HEX_ERR_READ)
			    tprintf(":LH failed: read error: %u\n", _hexfile_stack.uart_errors);
            else
                tprintf(":LH failed: hex format\n");
            con_mgr_inq_enable();
			xbank_free_program_lock();
			return;
		}
		lines++;
		
		switch(_hexfile_stack.record.type){
			case HEX_RECORD_DATA:
				if(start_new_record){
					DEBUG("new record: addr=%lx\n", base_addr+_hexfile_stack.record.addr);
					retval = xbank_write_prog_record(base_addr+_hexfile_stack.record.addr);
					start_new_record = 0;
				}
				
				// write data
				retval = xbank_write_prog_data(_hexfile_stack.record.data, _hexfile_stack.record.datalen);
				break;
			case HEX_RECORD_EOR:
				DEBUG("eor\n");
				// generate data crc
				retval = xbank_write_prog_crc(NULL);
				terminate = 1;
				break;
			case HEX_RECORD_ESA:
				base_addr = ((u_long)_hexfile_stack.record.data[0] << 12) | ((u_long)_hexfile_stack.record.data[1] << 4);
				start_new_record = 1;
				DEBUG("esa=%lx\n", base_addr);
				break;
			case HEX_RECORD_ELA:
				base_addr = ((u_long)_hexfile_stack.record.data[0] << 24) | ((u_long)_hexfile_stack.record.data[1] << 16);
				start_new_record = 1;
				DEBUG("ela=%lx\n", base_addr);
				break;
			case HEX_RECORD_SSA:
				seg_start_addr = (((u_long)_hexfile_stack.record.data[0] << 12) | 
				                 ((u_long)_hexfile_stack.record.data[1] << 4)) + // ext. seg. addr part
				                 (((u_long)_hexfile_stack.record.data[2] << 8) |
				                 ((u_long)_hexfile_stack.record.data[3]));       // offset addr part
				DEBUG("ssa=%lx\n", seg_start_addr);
				break;
			default:
				tprintf(":LH failed: record type\n");
				xbank_free_program_lock();
                con_mgr_inq_enable();
				return;
		}
        
        if(retval){
            tprintf(":LH failed: program size\n");
            xbank_free_program_lock();
            con_mgr_inq_enable();
            return;
        }		
	}                	
	
    
    // start addr
    xbank_set_prog_boot_addr(seg_start_addr);
	
	// check program
    if(xbank_validate_program()){
    	tprintf(":LH failed: program check\n");
    }else{
    	tprintf(":LH completed: %u lines read\n", lines);
    }
    
    con_mgr_inq_enable();
    xbank_free_program_lock();
}

void hexfile_init(FILE* stream){
    _hexfile_stack.stream = stream;
}

void hexfile_register_cmds(void)
{
    btn_terminal_register_cmd("loadhex", _loadhex_cmd);		
}
