/*
 * Copyright (C) 2000-2006 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * @file con_power_mgr.c
 *
 * @brief Implementation of a simple power manager
 *        for already established bluetooth connections
 *  
 * \author Mustafa Yuecel <yuecelm@ee.ethz.ch>
 * 
 * \date 2006/03/29
 *
 * $Log: con_power_mgr.c,v $
 * Revision 1.2  2006/04/06 17:04:46  yuecelm
 * some refactoring, reduced code size
 *
 * Revision 1.1  2006/04/06 13:55:53  yuecelm
 * add the connection power manager
 *
 */

#include <bt/bt_hci_api.h>
#include <bt/bt_hci_cmds.h>
#include <bt/bt_hci_defs.h>
#include <sys/event.h>
#include <sys/heap.h>
#include <sys/thread.h>
#include <terminal/eeprom_cmds.h>

// set default debug level
#define LOG_LEVEL 4
#define LOG_CLASS 5
#include <debug/log_set.h>

// EEPROM parameters
#define READ_EEPROM 0

#define PARAMS(y) void _cpm_eeprom_params_init(void) {y}
#define PARAMS_INIT() _cpm_eeprom_params_init()

#if READ_EEPROM > 0
#define EEPROM_PARAM(param_name, param_id, param_val, param_default_val) \
    eeprom_cmds_param_read_ext(param_id, &param_val, param_default_val); \
    eeprom_cmds_add_param(param_name, param_id, &param_val, param_default_val);
#else  // READ_EEPROM < 1
#define EEPROM_PARAM(param_name, param_id, param_val, param_default_val) \
    param_val = param_default_val;
#endif // READ_EEPROM

// put 4. link to sniff -> no event returned from Zeevo
#define CPM_DEFAULT_MAX_SNIFF_LINKS 3

#define CPM_DEFAULT_THREAD_SLEEP_TIME 5000
#define CPM_DEFAULT_THREAD_WAIT_TIME  100
#define CPM_DEFAULT_THREAD_PRIORITY   100

#define CPM_DEFAULT_SNIFF_INTERVAL 512
#define CPM_DEFAULT_SNIFF_ATTEMPT  8
#define CPM_DEFAULT_SNIFF_TIMEOUT  8

/// Connection Power Manager CPM stack
typedef struct _cpm_stack {
    struct btstack * bt_stack;  // pointer to the bluetooth stack
    u_long cod;  // filter this CoD (0 to disable)
    HANDLE start;  // start event
    HANDLE stop;  // stop event
    u_char max_sniff_links;  // max. sniffed links (for slaves!)
    u_char thread_priority;  // thread priority
    u_long thread_wait_time;  // sleep time between connections
    u_long thread_sleep_time;  // sleep time after go through all connection 
    u_short sniff_interval;  // sniff interval SI
    u_short sniff_attempt;  // sniff attempt SA
    u_short sniff_timeout;  // sniff timeout ST
} cpm_stack_t;

/// Pointer to the CPM Stack
static cpm_stack_t* cpm_stack;

PARAMS(
    EEPROM_PARAM("cpm_msl", "pmsl", cpm_stack->max_sniff_links, CPM_DEFAULT_MAX_SNIFF_LINKS)
    EEPROM_PARAM("cpm_tst", "pmts", cpm_stack->thread_sleep_time, CPM_DEFAULT_THREAD_SLEEP_TIME)
    EEPROM_PARAM("cpm_twt", "pmtw", cpm_stack->thread_wait_time, CPM_DEFAULT_THREAD_WAIT_TIME)
    EEPROM_PARAM("cpm_tp", "pmtp", cpm_stack->thread_priority, CPM_DEFAULT_THREAD_PRIORITY)
    EEPROM_PARAM("cpm_si", "pmsi", cpm_stack->sniff_interval, CPM_DEFAULT_SNIFF_INTERVAL)
    EEPROM_PARAM("cpm_sa", "pmsa", cpm_stack->sniff_attempt, CPM_DEFAULT_SNIFF_ATTEMPT)
    EEPROM_PARAM("cpm_st", "pmst", cpm_stack->sniff_timeout, CPM_DEFAULT_SNIFF_TIMEOUT)
)

void cpm_start(void)
{
    NutEventPost(&cpm_stack->start);
}

void cpm_stop(void)
{
    NutEventPost(&cpm_stack->start);
}

char _check_con_handle(bt_hci_con_handle_t con_handle, char mode, char role)
{
    long res;
    
    // check mode
    res = bt_hci_local_mode_discovery(cpm_stack->bt_stack, con_handle);
    if ( res != BT_ERR_NO_CON && (res & 0xff) != mode )
    {
        return 1;
    }
    
    if (role == -1)
    {
        return 0;   
    }
    
    // check role
    res = bt_hci_local_role_discovery(cpm_stack->bt_stack, con_handle);
    if ( res != BT_ERR_NO_CON && res != role )
    {
        return 2;
    }
    
    if (cpm_stack->cod == 0)
    {
        return 0;
    }
        
    // TODO check service code
    //res = bt_hci_local_code_discovery(bt_stack, con_handles[i]);
    
    return 0;
}

void _cpm_exit_sniff_mode(void)
{
    bt_hci_con_handle_array con_handles;
    u_char nr_open_con, i;
    long res;
    
    INFO("stop putting connections to sniff,"
         " now put all connections to active\n");
    
    nr_open_con = (u_char) bt_hci_get_con_handles(cpm_stack->bt_stack,
                                                  con_handles);
    DEBUG("open connections: %d\n", nr_open_con);
    
    for (i=0; i<nr_open_con; i++)
    {
        // only proceed connections in sniff mode
        if (_check_con_handle(con_handles[i], BT_HCI_CONN_MODE_SNIFF, -1))
        {
            continue;
        }

        res = bt_hci_exit_sniff_mode(cpm_stack->bt_stack, BT_HCI_SYNC,
                                     con_handles[i]);
        if (res < 0)
        {
            WARNING("not possible to put handle %d to active (error %d)\n",
                    con_handles[i], -res);
        }
        else
        {
            INFO("handle %d is in active mode\n", con_handles[i]);
        }
    }    
}

THREAD(CPM, arg)
{
    bt_hci_con_handle_array con_handles;
    u_char nr_open_con, i, stopped, sniff_links;
    long res;
    
    NutThreadSetPriority(cpm_stack->thread_priority);
    
    while (1) // thread-loop
    {
        stopped = 0;
        
        // wait until event CPM_START released
        NutEventWait(&cpm_stack->start, NUT_WAIT_INFINITE);
        
        INFO("Start putting connections to sniff\n");
        
        while (1) // sniff-loop
        {
            // try to put all connections in active mode to sniff mode

            sniff_links = 0;

            nr_open_con = (u_char) bt_hci_get_con_handles(cpm_stack->bt_stack,
                                                          con_handles);
            for (i=0; i<nr_open_con; i++)
            {
                res = bt_hci_local_mode_discovery(cpm_stack->bt_stack,
                                                  con_handles[i]);
                if ( res != BT_ERR_NO_CON &&
                     (res & 0xff) == BT_HCI_CONN_MODE_SNIFF )
                {
                    sniff_links++;
                }
            }
            
            INFO("open connections: %d, sniff connections: %d\n",
                 nr_open_con, sniff_links);

            for (i=0; i<nr_open_con; i++)
            {
                if (sniff_links >= cpm_stack->max_sniff_links)
                {
                    INFO("Maximum of sniff links reached\n");
                    break;
                }
                
                // only proceed connections in active mode and in master role
                if (_check_con_handle(con_handles[i], BT_HCI_CONN_MODE_ACTIVE, BT_HCI_MY_ROLE_MASTER))
                {
                    continue;
                }

                DEBUG("wait %u ms, watch out stop event\n",
                      cpm_stack->thread_wait_time);
                if (NutEventWait(&cpm_stack->stop, cpm_stack->thread_wait_time)
                        == 0)  // if event released
                {
                    stopped = 1;
                    break;
                }

                res = bt_hci_sniff_mode(cpm_stack->bt_stack, BT_HCI_SYNC,
                                        con_handles[i],
                                        cpm_stack->sniff_interval,
                                        cpm_stack->sniff_interval,
                                        cpm_stack->sniff_attempt,
                                        cpm_stack->sniff_timeout);
                
                if (res < 0)
                {
                    DEBUG("not possible to put con_handle %d to sniff (error %d)\n",
                          con_handles[i], -res);
                }
                else
                {
                    if ((res & 0xff) == BT_HCI_CONN_MODE_SNIFF)
                    {
                        INFO("handle %d is in sniff mode (%d, %d, %d)\n",
                             con_handles[i], (u_short) (res >> 8) & 0xffff,
                             cpm_stack->sniff_attempt, cpm_stack->sniff_timeout);
                        sniff_links++;
                    }
                    else
                    {
                        // should never happen??
                        WARNING("strange mode change\n");
                    }
                }
            }
            
            DEBUG("sleep %u ms, watch out stop event\n", cpm_stack->thread_sleep_time);
            
            if (stopped || (NutEventWait(&cpm_stack->stop,
                                         cpm_stack->thread_sleep_time) == 0)) // if event released
            {
                _cpm_exit_sniff_mode();
                break; // sniff-loop
            }
        } // sniff-loop
    } // thread-loop
}

char cpm_init(struct btstack * bt_stack, u_long cod)
{
    long res;
       
    // check default link policy
    // its really not my task to set the right default link policies, 
    // but check and print out message is ok
    res = bt_hci_read_default_link_policy_settings(bt_stack, BT_HCI_SYNC);
    if (res < 0)
    {
        ERROR("cannot read default link policy (%lu)\n", -res);
        return -1;
    }
    else if (!(res & BT_HCI_LINK_POLICY_SNIFF_MODE))
    {
        ERROR("sniff mode not in default link policy\n");
        return -2;
    }
    
    //dyn alloc cm stack
    cpm_stack = NutHeapAlloc(sizeof(cpm_stack_t));
    
    cpm_stack->bt_stack = bt_stack;
    cpm_stack->cod = cod;
    
    cpm_stack->start = 0;
    cpm_stack->stop = 0;
    
    PARAMS_INIT();
    
    if ( ! NutThreadCreate("CPM", CPM, 0, 256) )
    {
        ERROR("no ressources available for thread allocation...\n");
        NutHeapFree(cpm_stack);
        return -4;
    }
    
    return 0;
}
