#include <string.h>
#include <debug/xrlog.h>
#include <sys/heap.h>
#include <hardware/ram_banks.h>
#include <stdio.h>


// Note: modulo operations would clearify the code, but would be inefficient!

// log ring buf
struct xrlog_s {
	u_long mem_start;
        ///< Start address of the buffer (in extended memory)
	u_short pos;
        ///< Current position in the buffer (most recent byte written)
	u_short size;		///< Number of currently buffered bytes
	u_short max_size;	///< Maximal nr. of bytes that can be buffered
};

inline u_short xrlog_get_first_byte(xrlog_t* xrlog)
{
        if (xrlog->size != xrlog->max_size) return 0;
        return xrlog->pos;
}

inline u_short xrlog_get_curr_pos(xrlog_t* xrlog)
{
        return xrlog->pos;
}


// writes data to the log ring buffer
short xrlog_write(xrlog_t* xrlog, u_char* buf, u_short len)
{
    // nr of remaining bytes ('til end of memspace has been reached)
    u_short rem_bytes;

    // don't write more data than buffer is able to handle
    if (len > xrlog->max_size)
		return XRLOG_ERR_INVALID_DATA_ACCESS;
    
    // adjust buffer size
    if (xrlog->size > xrlog->max_size - len)
        xrlog->size = xrlog->max_size;
    else
        xrlog->size += len;

    // if data cannot be written contiguously, we write 'til start of memspace
    // has been reached
    rem_bytes = xrlog->max_size - xrlog->pos;
    if (len > rem_bytes) {
        cpy_to_xbank(buf, xrlog->mem_start + xrlog->pos, rem_bytes);
        // do wrap-around of position pointer
        xrlog->pos = 0;
        len -= rem_bytes;
        buf += rem_bytes;
    }
    // write (remaining) data
    cpy_to_xbank(buf, xrlog->mem_start + xrlog->pos, len);
    xrlog->pos += len;
    return 0;
}


/**
 * \brief Reads data from the ring buffer log.
 * 
 * Reads \param len bytes from the ring buffer. Start position is given by
 * the data pointer \param rbuf_p passed.
 * 
 * \return memory location (offset) until where data has been read
 * 
 */
short xrlog_read(xrlog_t* xrlog, u_char* buf, u_short* pos, u_short len)
{
    // nr of remaining bytes ('til end of memspace has been reached)
    u_short rem_bytes;

    // check if handle is valid
    if ((*pos) == xrlog->pos) return XRLOG_ERR_INVALID_DATA_ACCESS;

    // check if read access is valid
    rem_bytes = xrlog->max_size - (*pos) + xrlog->pos;
    if (rem_bytes > xrlog->max_size) rem_bytes -= xrlog->max_size;
    if (rem_bytes < len) return XRLOG_ERR_INVALID_DATA_ACCESS;

    // do wrap-around if necessary
    rem_bytes = xrlog->max_size - (*pos);
    if (rem_bytes < len) {
        // read data 'til the end of the memspace
        cpy_from_xbank(buf, xrlog->mem_start + (*pos), rem_bytes);
        // adjust remainin sizes & do wrap-around
        buf += rem_bytes;
        len -= rem_bytes;
        (*pos) = 0;
    }
    // read (remaining) data & adjust handle
    cpy_from_xbank(buf, xrlog->mem_start + (*pos), len);
    (*pos) += len;
    return 0;
}

short xrlog_seek(xrlog_t* xrlog, u_short* pos, short offset)
{
    // nr of remaining bytes
    u_short rem_bytes;

    if (offset >= 0) {
	    // check if handle is valid
	    if ((*pos) == xrlog->pos) return XRLOG_ERR_INVALID_DATA_ACCESS;
    	
        // check if data access is valid
        rem_bytes = xrlog->max_size - (*pos) + xrlog->pos;
        if (rem_bytes > xrlog->max_size) rem_bytes -= xrlog->max_size;
        if (rem_bytes < offset) return XRLOG_ERR_INVALID_DATA_ACCESS;

        // do wrap-around if necessary
        rem_bytes = xrlog->max_size - (*pos);
        if (rem_bytes < offset) {
            offset -= rem_bytes;
            (*pos) = 0;
        }
        //read (remaining) data & adjust handle
        (*pos) += offset;
    }
    else {
        // else, do sign conversion
        offset = -offset;

        // return if data access is not valid
        rem_bytes = xrlog->size - xrlog->pos + (*pos);
        if ((*pos) > xrlog->pos) rem_bytes -= xrlog->max_size;
        if (rem_bytes < offset) return XRLOG_ERR_INVALID_DATA_ACCESS;
        
        // else, do wrap-around if necessary
        if (offset > (*pos)) {
            offset -= (*pos);
            (*pos) = xrlog->max_size;
        }
        // seek
        (*pos) -= offset;
    }
    return 0;
}

    

void xrlog_empty(xrlog_t* xrlog)
{    
	xrlog->pos = 0;
	xrlog->size = 0;
}


xrlog_t* xrlog_init(u_long mem_addr, u_short max_size)
{
	xrlog_t* xrlog;
	
	// dyn. allocate ring buffer
	xrlog = NutHeapAllocClear(sizeof(xrlog_t));
	xrlog->mem_start = mem_addr;
	xrlog->max_size = max_size;
	xrlog->pos = 0;
	xrlog->size = 0;
	return xrlog;
}
