/*
 * Copyright (C) 2000-2005 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * \file wsnlab/wsnlab.c
 *
 * \date 16.05.2006
 *
 * \author bo
 *
 * \brief Quick hack library for the WSNLab2006
 *
 */
 
#ifndef _WSNLAB_H_
#define _WSNLAB_H_

// ----------------------------------------------------------------------------
// structs & typedefs
// ----------------------------------------------------------------------------

// forward declaration (does not work)
//typedef struct ccc_packet_st ccc_packet_t;

#include <cc/ccc.h>

typedef struct pkt_element
{
	u_char send_counter;
	u_char recv_counter;
	u_long received_at;
	ccc_packet_t* pkt;
} pkt_element_t;
 
 
// ----------------------------------------------------------------------------
// helper functions
// ----------------------------------------------------------------------------

/**
 * Prints the contents of a packet to the standard output.
 */
void print_packet_short( ccc_packet_t* pkt ); 
 
/**
 * Prints a string representation of the data to stdout (like a hex editor)
 */
void print_data( char* data, u_short length );

/**
 * Computes 2^e
 * 
 * \param e Exponent
 * \return 2^e
 */
u_short pow2(u_short e);

/**
 * Sets the LEDs according to a binary value 0..15
 * 
 * \param i Pattern to set (0..15)
 */
void led_set_binary(u_short i);

/**
 * Sends a packet, using the ccc interface.
 * 
 * \param dest Destination address.
 * \param type Type of the packet.
 * \param data Pointer to the data to send.
 * \param length Length of the data to send.
 * \return ???
 */
int ccc_send_packet( u_short dest, u_char type, u_char* data, u_short length );

/**
 * Beeps for one second, using the BTSense beeper.
 */
void beep(void);

// ----------------------------------------------------------------------------
// Dictionary
// ----------------------------------------------------------------------------

/**
 * Initializes the dictionary
 */
void dict_init(void);

/**
 * Tests if a packet is known, i.e. if it is in the dictionary.
 * 
 * \param pkt packet to test
 * \return 1 if the packet is known, 0 otherwise
 */
int dict_packet_is_known( ccc_packet_t *pkt );


/**
 * Adds an element to the packet dictionary, if there is space.
 * 
 * \param pkt The packet to add.
 * \return The index where the element has placed. If it equals 
 *         DICT_LENGTH, the element has not been added.
 */
int dict_add_packet( ccc_packet_t *pkt );

/**
 * Removes a packet with the handle idx from the dictionary. 
 * 
 * \param idx Handle of the packet.
 */
void dict_remove_packet( int idx );

/**
 * Returns the first Element of the dictionary
 * 
 * \param handle This will be set and is required for 
 * subsequent calls to dict_get_next_packet()
 */
pkt_element_t* dict_get_first_packet( int* handle );

/**
 * Iterates over the elements in the dictionary.
 * 
 * \param handle This will be set and is required for 
 * subsequent calls to dict_get_next_packet()
 */
pkt_element_t* dict_get_next_packet( int* handle );

/**
 * Prints the contents of the dictioary to the standard output
 */
void dict_print(void);

// ----------------------------------------------------------------------------
// ADC Stuff
// ----------------------------------------------------------------------------

enum adc2_ref_type
{
    ADC2_REF_AREF = 0,
    ADC2_REF_AVCC = 1 << 6,
    ADC2_REF_INTERNAL_256 = 3 << 6
};
typedef enum adc2_ref_type adc2_ref_t;

enum adc2_mode_type
{
    ADC2_MODE_SINGLE_CONVERSION = 0,
    ADC2_MODE_FREE_RUNNING      = 1 << 5,
};
typedef enum adc2_mode_type adc2_mode_t;

enum adc2_interrupt_type
{
    ADC2_INTERRUPT_DISABLE = 0,
    ADC2_INTERRUPT_ENABLE  = 1 << 3
};
typedef enum adc2_interrupt_type adc2_interrupt_t;

enum adc2_prescaler_type
{
	ADC2_PRESCALE_DIV2   = 0x00,    ///< 0x01,0x00 -> CPU clk/2
	ADC2_PRESCALE_DIV4   = 0x02,    ///< 0x02 -> CPU clk/4
	ADC2_PRESCALE_DIV8   = 0x03,    ///< 0x03 -> CPU clk/8
	ADC2_PRESCALE_DIV16  = 0x04,    ///< 0x04 -> CPU clk/16
	ADC2_PRESCALE_DIV32  = 0x05,    ///< 0x05 -> CPU clk/32
	ADC2_PRESCALE_DIV64  = 0x06,    ///< 0x06 -> CPU clk/64
	ADC2_PRESCALE_DIV128 = 0x07     ///< 0x07 -> CPU clk/128	
};
typedef enum adc2_prescaler_type adc2_prescaler_t;

enum adc2_channel_type
{
    ADC2_CHANNEL_0 = 0,
    ADC2_CHANNEL_1 = 1,
    ADC2_CHANNEL_2 = 2,
    ADC2_CHANNEL_3 = 3,
    ADC2_CHANNEL_4 = 4,
    ADC2_CHANNEL_5 = 5,
    ADC2_CHANNEL_6 = 6,
    ADC2_CHANNEL_7 = 7
};
typedef enum adc2_channel_type adc2_channel_t;

u_short adc2_init( adc2_mode_t mode, 
				adc2_interrupt_t use_interrupt, 
				adc2_prescaler_t prescaler, 
				adc2_channel_t channel, 
				adc2_ref_t reference );

u_short adc2_read( u_short handle );

#endif
